import itertools
import yaml
import os
from pathlib import Path, PosixPath as _PosixPath_, WindowsPath  as _WindowsPath_


class RPath(Path) :
    def __new__(cls, *args, **kvps):
        return super().__new__(WindowsPath if os.name == 'nt' else PosixPath, *args, **kvps)

    @classmethod
    def texnew(cls):
        return cls.home() / '.texnew'

    @classmethod
    def workspace(cls):
        return cls.home() / '.texnew' / '.workspace'

    @classmethod
    def templates(cls):
        return cls.home() / '.texnew' / 'templates'

    def read_yaml(self):
        return yaml.safe_load(self.read_text())

    def safe_rename(self):
        """Safely rename a file pointed to by path"""
        for t in itertools.count():
            new_path = self.with_name("{}_{}".format(self.stem, t) + "".join(self.suffixes))
            if not new_path.exists():
                self.rename(new_path)
                break

    @staticmethod
    def clean_workspace():
        """Deletes all files in '~/.texnew/.workspace' except .gitignore"""
        for p in RPath.workspace().iterdir():
            if not p.name == ".gitignore":
                p.unlink()


class WindowsPath(_WindowsPath_, RPath) :
    pass


class PosixPath(_PosixPath_, RPath) :
    pass



"""
Entry point for the texnew script installed along with this package.



    texnew [-uv] target template

    texnew [-cl]
    texnew -h | --help
    texnew -V | --version

Positional:
  target                The name of the file to action.
  template              The name of the template to use.

Optional:
  -u                    Update the target template with the given file.
  --keep-formatting     When updating, preserve the formatting section in the file.
  -v                    Run verbose.

Other:
  -c                    Check existing templates for errors.
  -l                    List existing templates.
"""
import sys
import argparse

from .scripts import run, run_update, run_check
from .template import available_templates
from .rpath import RPath
from . import __version__

def tn(args):
    if not args.output[0].endswith(".tex"):
        args.output[0] += ".tex"
    run(args.output[0], args.template[0])

def tn_update(args):
    if args.format:
        args.transfer.extend(['formatting','doctype'])
    run_update(args.target[0], args.template[0], transfer=args.transfer)

def tn_check(args):
    run_check(args.names,run_all=args.all)

def tn_info(args):
    # TODO
    # point to installation location
    # have listing option
    msg = {'lst':"Existing templates:\n"+ "\t".join(available_templates()),
            'dir':'Root directory: {}'.format(RPath.texnew())}
    d = vars(args)
    for k in msg.keys():
        if d[k]:
            print(msg[k])


def parse_errors(args):
    checklist = {'texnew_exists': (RPath.texnew().exists(), "Missing template information at '{}'".format(RPath.texnew())),
            'texnew_workspace_exists': (RPath.workspace().exists(), "Missing workspace directory at '{}'".format(RPath.workspace()))}
    for k,(check,val) in checklist.items():
        if not check:
            print("Error: ", val)

def parse():
    """Main argument parser"""
    parser = argparse.ArgumentParser(prog="texnew",description='An automatic LaTeX template creator and manager.')
    parser.add_argument('-v','--verbose',
            action='store_true',
            default=False,
            dest='verbose',
            help='run verbose mode')
    parser.add_argument('--version',
            action='version',
            version='%(prog)s {}'.format(__version__))
    subparsers = parser.add_subparsers(help="test")

    # main arguments
    parser_main = subparsers.add_parser('new', help='create a new template')
    parser_main.set_defaults(func=tn)
    parser_main.add_argument('output',
            type=str,
            nargs=1,
            help='the name of the file you want to create')
    parser_main.add_argument('template',
            type=str,
            nargs=1,
            help='the name of the template to use')

    # update arguments
    parser_update = subparsers.add_parser('update', help='update an existing file')
    parser_update.set_defaults(func=tn_update)
    parser_update.add_argument('target',
            type=str,
            nargs=1,
            help='the name of the file to update')
    parser_update.add_argument('template',
            type=str,
            nargs=1,
            help='the name of the template to update with')
    parser_update.add_argument("-f","--keep-formatting",
            dest="format",
            action="store_true",
            default=False,
            help="preserve the formatting block")
    parser_update.add_argument("-t","--transfer",
            nargs="*",
            default=['file-specific preamble', 'main document'],
            help="provide a list of blocks to transfer")

    # check arguments
    parser_check = subparsers.add_parser('check', help='check templates for errors')
    parser_check.set_defaults(func=tn_check)
    parser_check.add_argument('names',
            nargs="*",
            default=False,
            help="check all existing templates for errors")
    parser_check.add_argument('-a', "--all",
            action="store_true",
            default=False,
            dest="all",
            help="check all existing templates for errors")

    # info arguments
    parser_info = subparsers.add_parser('info', help='print information about the parser')
    parser_info.set_defaults(func=tn_info)
    parser_info.add_argument('-l', "--list",
            action="store_true",
            default=False,
            dest="lst",
            help="list existing templates")
    parser_info.add_argument('-d', "--directory",
            action="store_true",
            default=False,
            dest="dir",
            help="display path to the root folder")

    args = parser.parse_args()
    parse_errors(args)

    args.func(args)



from .rpath import RPath
from .document import TexnewDocument


def available_templates():
    """Print available templates"""
    return [s.stem for s in RPath.templates().iterdir()]


def load_template(template_type):
    """Load template information for template_data"""
    return (RPath.templates() / (template_type + '.yaml')).read_yaml()


def load_user(order=['private','default']):
    """Load user information for sub_list"""
    for path in [(RPath.texnew() / 'user' / (a+".yaml")) for a in order]:
        if path.exists():
            return path.read_yaml()
    raise FileNotFoundError('Could not find user file!')


# TODO: option to load custom defaults list from template (maybe default list should be found in the root of the template set)
def build(template_data, defaults=['doctype','packages','macros']):
    """Build a TexnewDocument from existing template_data."""
    tdoc = TexnewDocument({}, sub_list=template_data['substitutions'])
    p = RPath.texnew() / 'share' / template_data['template']

    # set default header
    tdoc['header'] = None

    # default components
    for name in defaults:
        tdoc[name] = (p / "defaults" / (name + ".tex")).read_text()

    # special macros
    for name in template_data['macros']:
        tdoc['macros ({})'.format(name)] = (p / "macros" / (name + ".tex")).read_text()
    
    # (space for) user preamble
    tdoc['file-specific preamble'] =  None

    # file constants
    constants = (p / "formatting" / (template_data['formatting']+ "_constants.tex"))
    if constants.exists():
        tdoc['constants'] = constants.read_text()
    else:
        tdoc['constants'] = None

    # formatting
    tdoc['formatting'] = (p / "formatting" / (template_data['formatting']+ ".tex")).read_text()

    # user space
    tdoc['main document'] = (p / "contents" / (template_data['contents']+ ".tex")).read_text()

    return tdoc


def update(tdoc, template_data, transfer):
    """Update a template document with a new template type, preserving the blocks specified in the 'transfer' list"""
    new_tdoc = build(template_data)

    # write information to new document
    for bname in transfer:
        new_tdoc[bname] = tdoc[bname]

    # transfer constants
    old_constants = tdoc.get_constants()
    new_tdoc.set_constants(old_constants)

    return new_tdoc
